import os
import aimrun
import click
import torch
import functools
import numpy as np
import torch.nn as nn
import torch.optim as optim
import torchvision
import torchvision.transforms as transforms
from mltiming import timing_iterator, timing
import random
from torch.utils.data import DataLoader
from torch.utils.data.distributed import DistributedSampler
from torch.distributed.fsdp import FullyShardedDataParallel as FSDP, MixedPrecision, ShardingStrategy
from torch.distributed.fsdp.wrap import transformer_auto_wrap_policy
from transformers import ViTForImageClassification, ViTConfig
from tqdm import tqdm
from torch.optim import AdamW
from torch.optim.lr_scheduler import StepLR
from detonation import DeMoReplicator, FullReplicator, NoReplicator, RandomReplicator, prepare_detonation
from transformers.models.vit.modeling_vit import ViTLayer
import torch.distributed as dist


@click.command()
@click.option('--dataset', default='cifar100', type=click.Choice(['cifar10', 'cifar100']))
@click.option('--batch-size', default=32, help='input batch size for training and validation (default: 32)')
@click.option('--epochs', default=10, help='number of epochs to train (default: 10)')
@click.option('--optim', default='deto-demo', type=click.Choice(['deto-demo', 'deto-full', 'deto-none', 'adamw', 'deto-random']))
@click.option('--compression-rate', default=0.1)
@click.option('--compression-topk', default=2)
@click.option('--compression-chunk', default=64)
@click.option('--replicate-every', default=1)
@click.option('--skip-every', default=None, type=int)
@click.option('--device', type=click.Choice(['cpu', 'cuda', 'mps']), default='cuda')
@click.option('--shards', default=None, type=int, help="Number of shards per replication group (default: number of GPUs per node)")
@click.option('--rand-seed', default=None, type=int, help="Seed for random generators in numpy and torch")
def main(dataset, batch_size, epochs, optim, compression_rate, compression_topk, compression_chunk, replicate_every, skip_every, device, shards, rand_seed):
    rank, nnodes, gpus = int(os.environ['RANK']), int(os.environ['NNODES']), int(os.environ['GPUS'])
    run_args = click.get_current_context().params
    run_args.update({
        'nnodes': nnodes,
        'gpus': gpus,
    })
    aimrun.init(repo='.', experiment='ViT', args=run_args)
    if rank == 0:
        print(aimrun.get_runs()[0].hash)
    single = device in ('cpu', 'mps') or (device == 'cuda' and nnodes == gpus == 1)
    model_and_co = setup(dataset, batch_size, optim, compression_rate, compression_topk, compression_chunk, replicate_every, skip_every, device, single, shards, rand_seed)
    train(epochs, optim, single, *model_and_co)

def seed(seed: int):
    random.seed(seed)
    np.random.seed(seed)
    torch.manual_seed(seed)
    if torch.cuda.is_available():
        torch.cuda.manual_seed_all(seed)
    elif torch.mps.is_available():
        torch.mps.manual_seed()

def train(epochs, optim, single, model, train_loader, val_loader, optimizer, scheduler, train_sampler):
    rank = int(os.environ['RANK'])
    for epoch in range(1, epochs+1):
        model.train()
        train_sampler.set_epoch(epoch)
        loss_samples = torch.zeros(2).to(model.device)
        metrics = {}
        for inputs, targets in tqdm(train_loader, desc=f"Training epoch {epoch}", disable=rank>0, colour="blue", ncols=150):
            if single:
                batch = batch.to(model.device)
            optimizer.zero_grad()
            if optim == 'adamw' or single:
                loss = model(inputs, labels=targets).loss
                loss.backward()
            else:
                with model.no_sync(): # Disable gradient replication for the backward pass
                    loss = model(inputs, labels=targets).loss
                    loss.backward()
            optimizer.step()
            loss_samples[0] += loss.item()
            loss_samples[1] += len(inputs)
            metrics.update({'train/loss': loss.item()})
            aimrun.track(metrics)
        # print training statistics
        if not single:
            dist.all_reduce(loss_samples, op=dist.ReduceOp.SUM)
        if rank == 0:
            train_loss = loss_samples[0] / loss_samples[1]
            print(f"Epoch {epoch} training loss  : {train_loss:.4f}")
            aimrun.track({'epoch/train/loss': train_loss}, step=epoch)
        # validation
        model.eval()
        loss_samples.zero_()
        correct = 0
        total = 0
        with torch.no_grad():
            for inputs, targets in tqdm(val_loader, desc=f"Validating after epoch {epoch}", disable=rank>0, colour="green", ncols=150):
                inputs, targets = inputs.to(model.device), targets.to(model.device)
                outputs = model(inputs, labels=targets)
                loss = outputs.loss
                loss_samples[0] += loss.item()
                loss_samples[1] += len(inputs)
                metrics.update({'val/loss': loss.item()})
                aimrun.track(metrics)
                _, predicted = outputs.logits.max(1)
                total += targets.size(0)
                correct += predicted.eq(targets).sum().item()
        # End of epoch - calculate global accuracy
        correct_tensor = torch.tensor(correct, device=model.device)
        total_tensor = torch.tensor(total, device=model.device)
        dist.all_reduce(total_tensor, op=dist.ReduceOp.SUM)
        global_train_acc = 100.0 * correct_tensor.item() / total_tensor.item()
        # print validation statistics
        if not single:
            dist.all_reduce(loss_samples, op=dist.ReduceOp.SUM)
        if rank == 0:
            val_loss = loss_samples[0] / loss_samples[1]
            print(f"Epoch {epoch} validation Loss: {val_loss:.4f}, Train Acc: {global_train_acc:.3f}%")
            aimrun.track({'epoch/val/loss': val_loss}, step=epoch)
            aimrun.track({'epoch/val/accuracy': global_train_acc}, step=epoch)
        scheduler.step()
    dist.destroy_process_group()
    aimrun.close()

def setup(dataset, batch_size, optim, compression_rate, compression_topk, compression_chunk, replicate_every, skip_every, device, single, shards, rand_seed):
    if rand_seed is not None:
        seed(rand_seed)

    config = ViTConfig(
        image_size=224,
        patch_size=16,
        num_channels=3,
        hidden_size=384,  # ViT-Small uses 384 dim
        num_hidden_layers=12,
        num_attention_heads=6,
        intermediate_size=384 * 4, # MLP size is typically 4x hidden size
        hidden_dropout_prob=0.1,
        attention_probs_dropout_prob=0.1,
        num_labels= 100 if dataset == 'cifar100' else 10
    )
    model = ViTForImageClassification(config)
    model = model.to(device)

    # Define Transforms
    transform_train = transforms.Compose([
        transforms.RandomCrop(32, padding=4),
        transforms.RandomHorizontalFlip(),
        transforms.Resize(224),  # Resize to ViT's expected input size
        transforms.ToTensor(),
        transforms.Normalize((0.4914, 0.4822, 0.4465), (0.2470, 0.2435, 0.2616)) if dataset == 'cifar10' else
        transforms.Normalize((0.5071, 0.4867, 0.4408), (0.2675, 0.2565, 0.2761))
    ])
    
    transform_test = transforms.Compose([
        transforms.Resize(224),  # Resize to ViT's expected input size
        transforms.ToTensor(),
        transforms.Normalize((0.4914, 0.4822, 0.4465), (0.2470, 0.2435, 0.2616)) if dataset == 'cifar10' else
        transforms.Normalize((0.5071, 0.4867, 0.4408), (0.2675, 0.2565, 0.2761))
    ])

    # Load dataset
    if dataset == 'cifar10':
        train_dataset = torchvision.datasets.CIFAR10(root='./data', train=True, download=True, transform=transform_train)
        val_dataset = torchvision.datasets.CIFAR10(root='./data', train=False, download=True, transform=transform_test)
    else:  
        train_dataset = torchvision.datasets.CIFAR100(root='./data', train=True, download=True, transform=transform_train)
        val_dataset = torchvision.datasets.CIFAR100(root='./data', train=False, download=True, transform=transform_test)
    
    train_sampler = DistributedSampler(train_dataset, shuffle=True)
    train_loader = DataLoader(train_dataset, batch_size=batch_size, shuffle=False, num_workers=4, sampler=train_sampler)
    val_loader = DataLoader(val_dataset, batch_size=batch_size, shuffle=False, num_workers=4, sampler=DistributedSampler(val_dataset))
    # prepare distributed training
    if device == 'cuda':
        torch.cuda.set_device(int(os.environ['LOCAL_RANK']))
    
    auto_wrap_policy = functools.partial(transformer_auto_wrap_policy, transformer_layer_cls={ViTLayer})
    mixed_precision = MixedPrecision(param_dtype=torch.bfloat16, reduce_dtype=torch.bfloat16, buffer_dtype=torch.bfloat16) if torch.cuda.is_bf16_supported() else None
    if single:
        model = model.to(device)
        optimizer = AdamW(model.parameters(), lr=1e-3, weight_decay=0.)
    elif optim.startswith('deto-'):
        if optim == 'deto-demo':
            replicator = DeMoReplicator(compression_topk=compression_topk, compression_chunk=compression_chunk)
        elif optim == 'deto-random':
            replicator = RandomReplicator(compression_rate=compression_rate, compression_chunk=compression_chunk, seed=rand_seed if rand_seed is not None else 42)
        elif optim == 'deto-full':
            replicator = FullReplicator()
        else:
            replicator = NoReplicator()
        model, optimizer = prepare_detonation(model, replicator, fsdp_kwargs={"auto_wrap_policy": auto_wrap_policy, "mixed_precision": mixed_precision}, replicate_every=replicate_every, skip_every=skip_every, sharding_group_size=shards)
    optim = optimizer._optimizer if hasattr(optimizer, "_optimizer") else optimizer
    scheduler = StepLR(optim, step_size=1, gamma=0.85)
    return model, train_loader, val_loader, optimizer, scheduler, train_sampler

if __name__ == '__main__':
    main()


from .demo_replicator import *
from .full_replicator import *
from .no_replicator import *
from .random_replicator import *
from .replicator import *
from .adamw_replicator import * 
from .slicing_replicator import *
from .striding_replicator import *



"""DeMo: Decoupled Momentum Optimization

This implements DeMo replication for DeToNATION.
"""

from .DeToSGD import DeToSGD
from ..repl import DeMoReplicator

__all__ = ["DeMo"]

class DeMo(DeToSGD):

    def __init__(
        self,
        params,
        compression_decay: float = 0.999,
        compression_topk: int = 32,
        compression_chunk: int = 64,
        **kwargs,
    ):
        super().__init__(
            params,
            replicator=DeMoReplicator(
                compression_decay=compression_decay,
                compression_topk=compression_topk,
                compression_chunk=compression_chunk,
            ),
            **kwargs,
        )



from einops import rearrange
import torch

__all__ = ["DCTCompress"]

class DCTCompress:

    def _clamp_topk(x, topk):
        if topk > x.shape[-1]:
            topk = x.shape[-1]
        if topk < 1:
            topk = 1
        return topk

    @classmethod
    @torch.no_grad()
    def compress(cls, x, topk):
        xshape = x.shape
        if len(x.shape) > 2:  # 2D weights
            x = rearrange(x, "y x h w -> y x (h w)")

        # Limit topk to max size
        topk = cls._clamp_topk(x, topk)

        idx = torch.topk(x.abs(), k=topk, dim=-1, largest=True, sorted=False).indices
        val = torch.gather(x, dim=-1, index=idx)

        return idx, val, xshape

    @torch.no_grad()
    def decompress(idx, val, xshape, device, dtype):
        x = torch.zeros(xshape, device=device, dtype=dtype)

        if len(xshape) > 2:  # 2D weights
            x = rearrange(x, "y x h w -> y x (h w)")

        # TODO: Careful, this is nondeterministic across different CUDA devices! might cause errors to accumulate between nodes!
        x.scatter_reduce_(dim=-1, index=idx, src=val, reduce="mean", include_self=False).reshape(xshape)

        if len(x.shape) > 2:  # 2D weights
            x = rearrange(x, "y x (h w) -> y x h w", h=xshape[2])

        return x

    @classmethod
    @torch.no_grad()
    def batch_decompress(cls, idx, val, xshape, device, dtype):
        idx = torch.concatenate(idx, dim=-1).to(device=device)
        val = torch.concatenate(val, dim=-1).to(device=device)
        return cls.decompress(idx, val, xshape, device, dtype)



import torch
import torch.distributed as dist
import torch.fft

from typing import Dict, Any

from .replicator import Replicator
from ..util import DCTCompress, DCTTransform
from typing import cast, List, Tuple
from torch import Tensor
from torch.optim.optimizer import _device_dtype_check_for_fused, _get_scalar_dtype

__all__ = ["AdamWDeMoReplicator"]

class AdamWDeMoReplicator(Replicator):

    def __init__(
        self,
        compression_decay: float = 0.999,
        compression_topk: int = 32,
        compression_chunk: int = 64,
    ):
        self.compression_decay = compression_decay
        self.compression_chunk = compression_chunk
        self.compression_topk = compression_topk

        if self.compression_topk <= 0:
            raise ValueError("topk_size has to be positive")
        if self.compression_chunk <= 0:
            raise ValueError("chunk_size has to be positive")
        if self.compression_decay < 0:
            raise ValueError("Negative compression_decay is currently not supported")
        if self.compression_decay >= 1:
            raise ValueError("Values of compression_decay bigger or equal to 1.0 is currently not supported")

    def init(
            self,
            optim: torch.optim.Optimizer,
            replication_parallel_group: dist.ProcessGroup | None = None,
        ):
        for group in optim.param_groups:
            amsgrad = group['amsgrad']
            for p in group["params"]:
                if p.requires_grad:
                    optim.state[p]["demo_delta"] = torch.zeros_like(p)

                state = optim.state[p]
                if group["fused"]:
                    _device_dtype_check_for_fused(p)
                # note(crcrpar): Deliberately host `step` on CPU if both capturable and fused are off.
                # This is because kernel launches are costly on CUDA and XLA.
                state["step"] = (
                    torch.zeros(
                        (),
                        dtype=_get_scalar_dtype(is_fused=group["fused"]),
                        device=p.device,
                    )
                    if group["capturable"] or group["fused"]
                    else torch.tensor(0.0, dtype=_get_scalar_dtype())
                )
                # Exponential moving average of gradient values
                state["exp_avg"] = torch.zeros_like(
                    p, memory_format=torch.preserve_format
                )
                # Exponential moving average of squared gradient values
                state["exp_avg_sq"] = torch.zeros_like(
                    p, memory_format=torch.preserve_format
                )
                if amsgrad:
                    # Maintains max of all exp. moving avg. of sq. grad. values
                    state["max_exp_avg_sq"] = torch.zeros_like(
                        p, memory_format=torch.preserve_format
                    )

        self.transform = DCTTransform(optim.param_groups, self.compression_chunk)
        self.replication_parallel_group = optim.replication_parallel_group if replication_parallel_group is None else replication_parallel_group
        self._replication_world_size = self.replication_parallel_group.size()
        self.data_transmitted = []
        self.data_received = []

    def pre_step(self):
        self.data_transmit = 0
        self.data_receive = 0

    def post_step(self):
        self.data_transmitted.append(self.data_transmit)
        self.data_received.append(self.data_receive)

    def replicate(
        self,
        sharded_grad: torch.Tensor,
        param: torch.nn.Parameter,
        param_state_dict: dict,
        param_group: Dict[str, Any],
    ) -> torch.Tensor:
        # Decay delta and add the gradient
        delta = param_state_dict["demo_delta"]
        if self.compression_decay != 1:
            delta.mul_(self.compression_decay)
        delta.add_(sharded_grad, alpha=param_group["lr"])

        # Replicating delta only if needed
        if self._replication_world_size == 1:
            new_grad = delta.clone()
            delta.zero_()
            return new_grad
    
        # Compress delta
        sparse_idx, sparse_val, xshape = DCTCompress.compress(
            self.transform.encode(delta), self.compression_topk
        )

        # Estimate transmitted delta
        transmit_grad = self.transform.decode(
            DCTCompress.decompress(sparse_idx, sparse_val, xshape, param.device, param.dtype)
        )

        # Remove transmitted from delta
        delta.sub_(transmit_grad)

        # Prepare and gather the indices and values
        sparse_idx_gather = [torch.zeros_like(sparse_idx) for _ in range(self._replication_world_size)]
        sparse_val_gather = [torch.zeros_like(sparse_val) for _ in range(self._replication_world_size)]
        sparse_idx_handle = dist.all_gather(sparse_idx_gather, sparse_idx, group=self.replication_parallel_group, async_op=True)
        sparse_val_handle = dist.all_gather(sparse_val_gather, sparse_val, group=self.replication_parallel_group, async_op=True)
        sparse_idx_handle.wait()
        sparse_val_handle.wait()

        # Log I/O data size
        self.data_transmit += sparse_idx.nbytes + sparse_val.nbytes
        for si, v in zip(sparse_idx_gather, sparse_val_gather):
            self.data_receive += si.nbytes + v.nbytes

        # Decode new gradient from all nodes
        new_grad = self.transform.decode(
            DCTCompress.batch_decompress(sparse_idx_gather, sparse_val_gather, xshape, param.device, param.dtype)
        )
        return new_grad



import torch
import torch.distributed as dist
from typing import Dict, Any

from .replicator import Replicator

__all__ = ["StridingReplicator"]

class StridingReplicator(Replicator):

    def __init__(
        self,
        compression_decay: float = 0.999,
        compression_rate: float = 0.1,
        compression_chunk: int = 64,
    ):
        self.compression_decay = compression_decay
        self.compression_rate = compression_rate
        self.compression_chunk = compression_chunk

        if self.compression_rate <= 0:
            raise ValueError("sample_factor has to be positive")
        if self.compression_rate > 1:
            raise ValueError("Values of compression_rate greater than 1.0 are not supported")

    def init(
            self,
            optim: torch.optim.Optimizer,
            replication_parallel_group: dist.ProcessGroup | None = None,
        ):
        self.counter = 0
        self.replication_parallel_group = optim.replication_parallel_group if replication_parallel_group is None else replication_parallel_group
        self._replication_world_size = self.replication_parallel_group.size()
        self.data_transmitted = []
        self.data_received = []
        for group in optim.param_groups:
            for p in group['params']:
                if p.requires_grad:
                    optim.state[p]["demo_delta"] = torch.zeros_like(p)

    def pre_step(self):
        self.data_transmit = 0
        self.data_receive = 0

    def post_step(self):
        self.data_transmitted.append(self.data_transmit)
        self.data_received.append(self.data_receive)

    def replicate(
        self,
        sharded_grad: torch.Tensor,
        param: torch.nn.Parameter,
        param_state_dict: dict,
        param_group: Dict[str, Any],
    ) -> torch.Tensor:
        # No replication if compression rate is 0
        if self.compression_rate == 0:
            return sharded_grad.to(device=param.device, dtype=param.dtype)

        # Decay delta and add the gradient
        delta = param_state_dict["demo_delta"]
        if self.compression_decay != 1:
            delta.mul_(self.compression_decay)
        delta.add_(sharded_grad, alpha=param_group["lr"])

        # Replicating delta only if needed
        if self._replication_world_size == 1 or self.compression_rate == 1:
            new_grad = delta.clone()
            delta.zero_()
            return new_grad

        # Compress delta
        closest_chunk = self.__class__._get_smaller_split(len(delta), self.compression_chunk)
        delta = delta.view(closest_chunk, -1)
        size = delta.size(0)
        chunks_per_step = int(max(self.compression_rate*size,1))
        stride = (size+chunks_per_step-1) // chunks_per_step
        offset = (self.counter * stride * chunks_per_step) % size
        indices = (offset + torch.arange(chunks_per_step, device=delta.device) * stride) % size
        compressed_grad = delta[indices]
        self.counter += 1

        # Remove compressed gradient from delta
        mask = torch.zeros(delta.size(0), dtype=torch.bool, device=param.device)
        mask[indices] = True
        delta.mul_(~mask.unsqueeze(1) if delta.dim() > 1 else ~mask)

        # Average the compressed gradient
        dist.all_reduce(compressed_grad, dist.ReduceOp.AVG, group=self.replication_parallel_group)

        # Log I/O data size
        self.data_transmit += compressed_grad.nbytes
        self.data_receive += compressed_grad.nbytes

        # Decode new gradient from all nodes
        new_grad = torch.zeros_like(delta, device=param.device)
        new_grad[indices] = compressed_grad
        new_grad = new_grad.view_as(sharded_grad)
        return new_grad

    def _get_prime_divisors(n):
        divisors = []
        while n % 2 == 0:
            divisors.append(2)
            n //= 2
        while n % 3 == 0:
            divisors.append(3)
            n //= 3
        i = 5
        while i * i <= n:
            for k in (i, i + 2):
                while n % k == 0:
                    divisors.append(k)
                    n //= k
            i += 6
        if n > 1:
            divisors.append(n)
        return divisors

    @classmethod
    def _get_divisors(cls, n):
        divisors = []
        if n == 1:
            divisors.append(1)
        elif n > 1:
            prime_factors = cls._get_prime_divisors(n)
            divisors = [1]
            last_prime = 0
            factor = 0
            slice_len = 0
            # Find all the products that are divisors of n
            for prime in prime_factors:
                if last_prime != prime:
                    slice_len = len(divisors)
                    factor = prime
                else:
                    factor *= prime
                for i in range(slice_len):
                    divisors.append(divisors[i] * factor)
                last_prime = prime
            divisors.sort()
        return divisors

    @classmethod
    def _get_smaller_split(cls, n, close_to):
        all_divisors = cls._get_divisors(n)
        for ix, val in enumerate(all_divisors):
            if val == close_to:
                return val
            if val > close_to:
                if ix == 0:
                    return val
                return all_divisors[ix - 1]
        return n



"""DeToNATION: Decoupled Torch Network-Aware Training on Interlinked Online Nodes

This implements the DeToNATION optimizer that works in DDP, FSDP, and HSDP settings. For gradient
synchronization in DDP and HSDP settings, it offers a number of approaches:
* no gradient replication
* Decoupled Momentum (DeMo) repliation - https://arxiv.org/abs/2106.11447 with the implementation based on https://github.com/bloc97/DeMo
* full gradient replication
In an exisiting codebase that uses PyTorch and one of these sharding strategies, wrap your forward-backward in 
`torch.distributed.DistributedDataParallel.no_sync` to disable external gradient synchronization.
See https://pytorch.org/docs/stable/generated/torch.nn.parallel.DistributedDataParallel.html#torch.nn.parallel.DistributedDataParallel.no_sync
TODO:
* auto-detect policy if none is given in prepare_detonation according to the model's structure
* compute transmitted data for full and no replication (and make that a part of the DeToNATION?)
* profiling of GPU, CPU, network, and wall time usage (optimize DeMoReplicator? stream replications instead of waiting?)
"""

import torch
import torch.distributed as dist
import torch.nn.functional as F
from typing import Callable, List

from .repl import DeMoReplicator, Replicator

__all__ = ["DeToNATIONMixin"]

class DeToNATIONMixin():

    def __init__(
        self,
        detonation_weight_decay: float = 0.0,
        detonation_sign: bool = True,
        sharding_parallel_group: dist.ProcessGroup | None = None,
        replication_parallel_group: dist.ProcessGroup | List[dist.ProcessGroup] | None = None,
        replicator: Replicator | List[Replicator] = DeMoReplicator(),
        replicate_every: int | List[int] = 1,
        skip_every: int | List[int] | None = None,
    ):
        self.detonation_weight_decay = detonation_weight_decay
        self.detonation_sign = detonation_sign
        self.sharding_parallel_group = sharding_parallel_group # intra-node communication 
        self.replication_parallel_groups = replication_parallel_group if isinstance(replication_parallel_group, list) else [replication_parallel_group] # inter-node communication
        self.replicators = replicator if isinstance(replicator, list) else [replicator]
        self.replicate_everys = replicate_every if isinstance(replicate_every, list) else [replicate_every]*len(self.replicators)
        self.skip_everys = [None]*len(self.replicators) if skip_every is None else (skip_every if isinstance(skip_every, list) else [skip_every])

        self._sharding_world_size = dist.get_world_size(self.sharding_parallel_group)
        if self._sharding_world_size == 0:
            raise ValueError("Sharding world size cannot be zero")
        for replication_parallel_group in self.replication_parallel_groups:
            if dist.get_world_size(replication_parallel_group) == 0:
                raise ValueError("Replication world size cannot be zero")
        self.state["detonation_step"] = 0
        for replicator, replication_parallel_group in zip(self.replicators, self.replication_parallel_groups):
            replicator.init(self, replication_parallel_group=replication_parallel_group)

    def _grad_reduce_scatter(self, grad: torch.Tensor):
        # Do not reduce_scatter if the gradient is not sharded
        if self._sharding_world_size == 1:
            return grad

        # Chunk and pad the unsharded gradient
        chunks = list(grad.chunk(self._sharding_world_size))
        numel_to_pad = self._sharding_world_size * chunks[0].numel() - grad.numel()
        padded_unsharded_grad = F.pad(grad, [0, numel_to_pad]) if numel_to_pad > 0 else grad

        # Prepare and scatter the sharded gradient
        sharded_grad = torch.empty_like(chunks[0])
        dist.reduce_scatter_tensor(
            sharded_grad,
            padded_unsharded_grad,
            op=dist.ReduceOp.AVG,
            group=self.sharding_parallel_group,
        )
        return sharded_grad

    def step(self, closure: Callable | None = None, base_step: torch.optim.Optimizer.step = None):
        self.state["detonation_step"] += 1

        # Any step-wise initialization needed by the replicator
        for replicator in self.replicators:
            replicator.pre_step()

        for group in self.param_groups:
            lr = group["lr"]
            for param in group["params"]:
                if not param.requires_grad:
                    continue

                # Sharding gradient if needed
                unsharded_grad = param.grad.data
                param.grad = None
                sharded_grad = self._grad_reduce_scatter(unsharded_grad)

                # Step-Weight decay
                if self.detonation_weight_decay != 0.0:
                    param.data.mul_(1.0 - lr * self.weight_decay)

                # Replicating the gradient if needed
                for replicate_every, skip_every, replicator in zip(self.replicate_everys, self.skip_everys, self.replicators):
                    if (
                        (self.state["detonation_step"] % replicate_every == 0) and
                        (skip_every is None or (self.state["detonation_step"] % skip_every != 0))
                    ):
                        new_grad = replicator.replicate(
                            sharded_grad=sharded_grad,
                            param=param,
                            param_state_dict=self.state[param],
                            param_group=group,
                        )
                    else:
                        new_grad = sharded_grad.to(param.device).to(param.dtype)
                param.grad = new_grad

                # Sign-SGD
                if self.detonation_sign:
                    param.grad.sign_()

        # SGD step
        result = base_step(self, closure)

        # Any step-wise finalization needed by the replicator
        for replicator in self.replicators:
            replicator.post_step()

        return result



from abc import ABC, abstractmethod
import torch
from typing import Any, Dict

__all__ = ["Replicator"]

class Replicator(ABC):
    """
    This provides the functions to initialize and execute replicators to be used by the
    :class:`DeToNATION` optimizer for replicating gradients across replication groups (typically nodes).
    """

    def init(
        self,
        optim: torch.optim.Optimizer,
        replication_parallel_group: torch.distributed.ProcessGroup | None = None,
    ):
        """
        Initialize the replicator.
        """
        pass

    def pre_step(self):
        """
        Pre-step hook.
        """
        pass

    def post_step(self):
        """
        Post-step hook.
        """
        pass

    @abstractmethod
    def replicate(
        self,
        sharded_grad: torch.Tensor,
        param: torch.nn.Parameter,
        param_state_dict: dict,
        param_group: Dict[str, Any],
    ):
        """
        Replicate the gradient across the replication group.

        Args:
            sharded_grad (torch.Tensor): The sharded gradient tensor.
            param (torch.nn.Parameter): The current parameter.
            param_state_dict (dict): The state dictionary of the parameter.
            param_group (Dict[str, Any]): The current parameter group.

        Returns:
            torch.Tensor: The replicated gradient tensor.
        """
        pass



import torch
from typing import Any, Dict

from .replicator import Replicator

__all__ = ["NoReplicator"]

class NoReplicator(Replicator):

    def replicate(
        self,
        sharded_grad: torch.Tensor,
        param: torch.nn.Parameter,
        param_state_dict: dict,
        param_group: Dict[str, Any],
    ) -> torch.Tensor:
        return sharded_grad.to(param.device).to(param.dtype)



from einops import rearrange
import math
import torch

__all__ = ["DCTTransform"]

class DCTTransform:

    @torch.no_grad()
    def __init__(self, param_groups, target_chunk, norm="ortho"):
        self.target_chunk = target_chunk

        self.shape_dict = dict()
        self.f_dict = dict()
        self.b_dict = dict()

        # Get all variants of model tensor sizes
        # Generate all possible valid DCT sizes for model tensors
        for group in param_groups:
            for p in group["params"]:
                if not p.requires_grad:
                    continue
                for s in p.shape:
                    # Get the closest smallest divisor to the targeted DCT size
                    sc = self.__class__._get_smaller_split(s, self.target_chunk)
                    self.shape_dict[s] = sc

                    # Pregenerate DCT basis matrices
                    if sc not in self.f_dict:
                        I = torch.eye(sc)
                        self.f_dict[sc] = self.__class__._dct(I, norm=norm).to(p.dtype).to(p.device)
                        self.b_dict[sc] = self.__class__._idct(I, norm=norm).to(p.dtype).to(p.device)

    @torch.no_grad()
    def _einsum_2d(x, b, d=None):
        if d is None:
            return torch.einsum("...ij, jb -> ...ib", x, b)
        else:
            # Note: b-c axis output is transposed to chunk DCT in 2D
            return torch.einsum("...ijkl, jb, ld -> ...ikbd", x, b, d)

    @torch.no_grad()
    def _einsum_2d_t(x, b, d=None):
        if d is None:
            return torch.einsum("...ij, jb -> ...ib", x, b)
        else:
            # Note: b-c axis output is transposed to chunk DCT in 2D
            return torch.einsum("...ijkl, kb, ld -> ...ibjd", x, b, d)

    @torch.no_grad()
    def encode(self, x):
        if len(x.shape) > 1:  # 2D weights
            n1 = self.shape_dict[x.shape[0]]
            n2 = self.shape_dict[x.shape[1]]
            n1w = self.f_dict[n1].to(x.device)
            n2w = self.f_dict[n2].to(x.device)
            self.f_dict[n1] = n1w
            self.f_dict[n2] = n2w

            x = rearrange(x, "(y h) (x w) -> y h x w", h=n1, w=n2)
            x = self.__class__._einsum_2d(x, n1w, n2w)

        else:  # 1D weights
            n1 = self.shape_dict[x.shape[0]]
            n1w = self.f_dict[n1].to(x.device)
            self.f_dict[n1] = n1w

            x = rearrange(x, "(x w) -> x w", w=n1)
            x = self.__class__._einsum_2d(x, n1w)

        return x

    @torch.no_grad()
    def decode(self, x):
        if len(x.shape) > 2:  # 2D weights
            n1 = x.shape[2]
            n2 = x.shape[3]
            n1w = self.b_dict[n1].to(x.device)
            n2w = self.b_dict[n2].to(x.device)
            self.b_dict[n1] = n1w
            self.b_dict[n2] = n2w

            x = self.__class__._einsum_2d_t(x, n1w, n2w)
            x = rearrange(x, "y h x w -> (y h) (x w)")

        else:  # 1D weights
            n1 = x.shape[1]
            n1w = self.b_dict[n1].to(x.device)
            self.b_dict[n1] = n1w

            x = self.__class__._einsum_2d_t(x, n1w)
            x = rearrange(x, "x w -> (x w)")

        return x

    # Code modified and sourced from https://github.com/zh217/torch-dct
    def _dct_fft_impl(v):
        return torch.view_as_real(torch.fft.fft(v, dim=1))

    def _idct_irfft_impl(V):
        return torch.fft.irfft(torch.view_as_complex(V), n=V.shape[1], dim=1)

    @classmethod
    def _dct(cls, x, norm=None):
        """
        Discrete Cosine Transform, Type II (a.k.a. the DCT)

        For the meaning of the parameter `norm`, see:
        https://docs.scipy.org/doc/scipy-0.14.0/reference/generated/scipy.fftpack.dct.html

        :param x: the input signal
        :param norm: the normalization, None or 'ortho'
        :return: the DCT-II of the signal over the last dimension
        """
        x_shape = x.shape
        N = x_shape[-1]
        x = x.contiguous().view(-1, N)

        v = torch.cat([x[:, ::2], x[:, 1::2].flip([1])], dim=1)

        Vc = cls._dct_fft_impl(v)

        k = -torch.arange(N, dtype=x.dtype, device=x.device)[None, :] * math.pi / (2 * N)
        W_r = torch.cos(k)
        W_i = torch.sin(k)

        V = Vc[:, :, 0] * W_r - Vc[:, :, 1] * W_i

        if norm == "ortho":
            V[:, 0] /= math.sqrt(N) * 2
            V[:, 1:] /= math.sqrt(N / 2) * 2

        V = 2 * V.view(*x_shape)

        return V

    @classmethod
    def _idct(cls, X, norm=None):
        """
        The inverse to DCT-II, which is a scaled Discrete Cosine Transform, Type III

        Our definition of idct is that idct(dct(x)) == x

        For the meaning of the parameter `norm`, see:
        https://docs.scipy.org/doc/scipy-0.14.0/reference/generated/scipy.fftpack.dct.html

        :param X: the input signal
        :param norm: the normalization, None or 'ortho'
        :return: the inverse DCT-II of the signal over the last dimension
        """

        x_shape = X.shape
        N = x_shape[-1]

        X_v = X.contiguous().view(-1, x_shape[-1]) / 2

        if norm == "ortho":
            X_v[:, 0] *= math.sqrt(N) * 2
            X_v[:, 1:] *= math.sqrt(N / 2) * 2

        k = torch.arange(x_shape[-1], dtype=X.dtype, device=X.device)[None, :] * math.pi / (2 * N)
        W_r = torch.cos(k)
        W_i = torch.sin(k)

        V_t_r = X_v
        V_t_i = torch.cat([X_v[:, :1] * 0, -X_v.flip([1])[:, :-1]], dim=1)

        V_r = V_t_r * W_r - V_t_i * W_i
        V_i = V_t_r * W_i + V_t_i * W_r

        V = torch.cat([V_r.unsqueeze(2), V_i.unsqueeze(2)], dim=2)

        v = cls._idct_irfft_impl(V)
        x = v.new_zeros(v.shape)
        x[:, ::2] += v[:, : N - (N // 2)]
        x[:, 1::2] += v.flip([1])[:, : N // 2]

        return x.view(*x_shape)

    def _get_prime_divisors(n):
        divisors = []
        while n % 2 == 0:
            divisors.append(2)
            n //= 2
        while n % 3 == 0:
            divisors.append(3)
            n //= 3
        i = 5
        while i * i <= n:
            for k in (i, i + 2):
                while n % k == 0:
                    divisors.append(k)
                    n //= k
            i += 6
        if n > 1:
            divisors.append(n)
        return divisors

    @classmethod
    def _get_divisors(cls, n):
        divisors = []
        if n == 1:
            divisors.append(1)
        elif n > 1:
            prime_factors = cls._get_prime_divisors(n)
            divisors = [1]
            last_prime = 0
            factor = 0
            slice_len = 0
            # Find all the products that are divisors of n
            for prime in prime_factors:
                if last_prime != prime:
                    slice_len = len(divisors)
                    factor = prime
                else:
                    factor *= prime
                for i in range(slice_len):
                    divisors.append(divisors[i] * factor)
                last_prime = prime
            divisors.sort()
        return divisors

    @classmethod
    def _get_smaller_split(cls, n, close_to):
        all_divisors = cls._get_divisors(n)
        for ix, val in enumerate(all_divisors):
            if val == close_to:
                return val
            if val > close_to:
                if ix == 0:
                    return val
                return all_divisors[ix - 1]
        return n
